%This function detect faces in an image

function face_detection(inputfile)
%create skin model from skin samples
[rmean,bmean,rbcov]=make_model();
%convert input image into skin likelyhood image
[likely_skin]=get_likelyhood(inputfile,rmean,bmean,rbcov);
%segmented skin by adaptive threshold method
[skinBW,opt_th] = segment_adaptive(likely_skin);
%apply erosion,dilations and label the regions
[erodedBW]=label_regions(skinBW);
%apply eular test i.e number holes >=1
[eulerBW]=euler_test(erodedBW);
%apply aspect ratio test
[aspectBW]=aspect_test(eulerBW);
%apply template matching test
[templateBW]=template_test(aspectBW,inputfile,'template.jpg');

[K,P]=bwlabel(templateBW,8);

%compute centroid of each region and plot on original image
s  = regionprops(bwlabel(templateBW), 'centroid');
centroids = cat(1, s.Centroid);
subplot(4,3,12);
imshow(imread(inputfile))
if(P>0)
hold on
plot(centroids(:,1), centroids(:,2), 'b*')
hold off
end
title('Final Detection')

clear all

%***********************************FUNCTIONS******************************
%**************************************************************************
% Following function returns cromatic values of an input image file
function [cr, cb] = get_crcb(filename)
im= imread(filename);
% convert RGB to YCbCr
imycc = rgb2ycbcr(im);
% low pass filter matrix
lpf = 1/9 * ones(3);
% take Cr and Cb channels
cr = imycc(:,:,3);
cb = imycc(:,:,2);
% pass thru low pass filter
cr = filter2(lpf, cr);
cb = filter2(lpf, cb);
%concatenate all rows
cr = reshape(cr, 1, prod(size(cr)));
cb = reshape(cb, 1, prod(size(cb)));

%**************************************************************************
%Following function plots 2D Chromatic Histogram
function plot_hist2d(cr, cb)
%creates 256 x 256 matrix of zeros - for quantized matrix
hist2d = zeros(256);
%round-off the crominance values into integers
cr = round(cr);
cb = round(cb);
%create 2D histogram function
for i = 1:length(cr)
   hist2d(cr(i),cb(i)) = hist2d(cr(i),cb(i))+1;
end
% %plot 2D histogram
surf(hist2d)

%**************************************************************************
%Following fnction fit the statistics of sample chrominance values into
%2D Gaussian model and returns the parameters of model
function [rmean,bmean,rbcov]=make_model()
%get crominance values of skin sample images
[cr1, cb1] = get_crcb('sampleset/1.jpg');
[cr2, cb2] = get_crcb('sampleset/2.jpg');
[cr3, cb3] = get_crcb('sampleset/3.jpg');
[cr4, cb4] = get_crcb('sampleset/4.jpg');
[cr5, cb5] = get_crcb('sampleset/5.jpg');
[cr6, cb6] = get_crcb('sampleset/6.jpg');
[cr7, cb7] = get_crcb('sampleset/7.jpg');
[cr8, cb8] = get_crcb('sampleset/8.jpg');
[cr9, cb9] = get_crcb('sampleset/9.jpg');
[cr10, cb10] = get_crcb('sampleset/10.jpg');
[cr11, cb11] = get_crcb('sampleset/11.jpg');
[cr12, cb12] = get_crcb('sampleset/12.jpg');
[cr13, cb13] = get_crcb('sampleset/13.jpg');
%concatenate all values
cr = [cr1 cr2 cr3 cr4 cr5 cr6 cr7 cr8 cr9 cr10 cr11 cr12 cr13];
cb = [cb1 cb2 cb3 cb4 cb5 cb6 cb7 cb8 cb9 cb10 cb11 cr12 cb13];
%compute statistics of sample values
rmean = mean(cr);
bmean = mean(cb);
rbcov = cov(cr,cb);
%The concept of linear transformation of random variables is applied 
%In this case, random variables Cr and Cb are jointly Gaussian
%Write joint distribution function
jointchart = zeros(256);
for r = 0:255
   for b = 0:255
		 x = [(r - rmean);(b - bmean)];
      jointchart(r+1,b+1) = [power(2*pi*power(det(rbcov),0.5),-1)]*exp(-0.5* x'*inv(rbcov)* x);
   end
end
%plot 2D histogram
subplot(4,3,1);
plot_hist2d(cr,cb)
title('2D Chromatic Histogram Model')
%plot joint distribution function
subplot(4,3,2);
surf(jointchart)
title('Joint Gaussian Distribution Model')


%**************************************************************************
%Following function convert original image into normalized skin likelyhood image
function[likely_skin]=get_likelyhood(filename,rmean,bmean,rbcov)
%read input file 
img = imread(filename);
% convert RGB to YCrCb color space
imycbcr = rgb2ycbcr(img);
[m,n,l] = size(img);
%create a 2D matrix with same dimension of image
likely_skin = zeros(m,n);
for i = 1:m
   for j = 1:n
      %get crominance values for each pixel
      cr = double(imycbcr(i,j,3));
      cb = double(imycbcr(i,j,2));
      %compute the likelyhood of each pixel
      x = [(cr-rmean);(cb-bmean)];
      likely_skin(i,j) = [power(2*pi*power(det(rbcov),0.5),-1)]*exp(-0.5* x'*inv(rbcov)* x);
   end
end
%pass thru low pass filter
lpf= 1/9*ones(3);
likely_skin = filter2(lpf,likely_skin);
%normalize the likelyhood values with maximum value
likely_skin = likely_skin./max(max(likely_skin));
%show skin likelyhood grayscale image
subplot(4,3,3);
imshow(img, [0 1])
title('Original RGB Image')
subplot(4,3,4);
imshow(likely_skin, [0 1])
title('Skin Likelyhood Image')

%**************************************************************************
%Following function take likelyhood image as input and segment it into binary image by
%setting the threshold adaptively
function [binary_skin,opt_th] = segment_adaptive(likely_skin)
%intialize
[m,n] = size(likely_skin);
temp = zeros(m,n);
diff_list = [];
%set threshold range and stepsize by experiment
high=0.55;
low=0.01;
step_size=-0.1;
bias_factor=1;
indx_count=[(high-low)/abs(step_size)]+2;
%finding optimal threshold
for threshold = high:step_size:low
   binary_skin = zeros(m,n);
   binary_skin(find(likely_skin>threshold)) = 1;
   diff = sum(sum(binary_skin - temp));
   diff_list = [diff_list diff];
   temp = binary_skin;
end
%optimal threshold is the threshold where minimum diff occur
[C, indx] = min(diff_list);
opt_th = (indx_count-indx)*abs(step_size)*bias_factor;
% reset the binary_skin image matrix
binary_skin = zeros(m,n);
binary_skin(find(likely_skin>opt_th)) = 1;
%show skin segmented binary image
subplot(4,3,5);
imshow(binary_skin, [0 1])
title('Skin Segmented Image')

%**************************************************************************
%Following function applies erosion to binary image and mark the regions
function[labelBW]=label_regions(binary_skin)
[m,n] = size(binary_skin); 
% %Apply Minimal Erosion to open up some small holes
% se0 = strel('disk',2); 
% erodedBW0=zeros(m,n);
% erodedBW0= imerode(binary_skin,se0);
%Fill the regions
filledBW=zeros(m,n);
filledBW = imfill(binary_skin,'holes');
%Apply Erosion
se2 = strel('disk',10); 
erodedBW=zeros(m,n);
erodedBW = imerode(filledBW,se2);
subplot(4,3,6);
imshow(erodedBW)
title({'After Erosion';'(disk size: 10)'})
%Apply Dilation
se1 = strel('disk',8); 
dilateBW=zeros(m,n);
dilateBW=imdilate(erodedBW,se1);
%multiply eroded image with skin segmented image to retain holes
dilateBW = immultiply(dilateBW,binary_skin);
subplot(4,3,7);
imshow(dilateBW)
title({'After Dilation';'(disk size: 8)'})
%Label skin regions
labelBW=zeros(m,n);
[labelBW,num] = bwlabel(dilateBW,8);
%Color the labeled regions -background is black
color_regions=zeros(m,n);
color_regions= label2rgb(labelBW, 'hsv', 'black', 'shuffle');
%show colored labeled regions of image
subplot(4,3,8);
imshow(color_regions)
title({'Labeled Regions';['(',num2str(num),' regions)']})

%**************************************************************************
%Following function filters regions which has at least one hole
function [eulerBW]=euler_test(labelBW)
%compute eular number of each region
e = regionprops(labelBW,'EulerNumber');
eulers=cat(1,e.EulerNumber);
%compute number of holes for each region
holes=1-eulers;
%take regions which has at least 1 hole
region_index = find(holes>=1);
[m,n]=size(labelBW);
eulerBW=zeros(m,n);
%make new binary image only with regions which pass euler test
for i=1:length(region_index)
    % Compute the coordinates for this region.
   [x,y] = find(bwlabel(labelBW) == region_index(i));
   % Get an image that only has this region, the rest is black 
   bwsegment = bwselect(labelBW,y,x,8);
   eulerBW=eulerBW+bwsegment;
end
subplot(4,3,9);
imshow(eulerBW)
title({'After Euler Test';['(',num2str(length(region_index)),' regions)']})

%**************************************************************************
%Following function determines whether the region aspect ratio is within
%range of being a face region
function [aspectBW]=aspect_test(eulerBW)
[m,n]=size(eulerBW);
%fill holes in image
filledBW = imfill(eulerBW,'holes');
%Apply Dilation
se1 = strel('disk',3); 
growBW=zeros(m,n);
growBW=imdilate(filledBW,se1);
%label as binary
[labels,num] = bwlabel(growBW,8);
[aspect_ratio]=get_aspect(labels);
% %take regions which has aspect ratio within range
region_index = find(aspect_ratio<=3.5 & aspect_ratio>=1);
aspectBW=zeros(m,n);

%make new binary image only with regions which pass euler test
for i=1:length(region_index)
    % Compute the coordinates for this region.
    [x,y] = find(bwlabel(filledBW) == region_index(i));
    % Get an image that only has this region, the rest is black 
    bwsegment = bwselect(filledBW,y,x,8);
    aspectBW=aspectBW+bwsegment;
end
subplot(4,3,10);
imshow(aspectBW)
title({'After Aspect Ratio Test';['(',num2str(length(region_index)),' regions)']})

%**************************************************************************
%Following function returns image regions which pass template matching test
function [template_passed]=template_test(aspectBW,originalRGB,template)

%convert original image into grayscale
imgray=rgb2gray(imread(originalRGB));
imtemplate=imread(template);

%label the binary image with regions
[labels,num] = bwlabel(aspectBW,8);
[m,n]=size(aspectBW);
% %compute orientation angle for each region
orient = regionprops(labels,'Orientation');
angles=cat(1,orient.Orientation);
%compute centroid for each region
c = regionprops(labels,'Centroid');
centroids=cat(1,c.Centroid);

%image with regions which pass template matching test
template_passed=zeros(m,n);
gray_matched=zeros(m,n);
%resize,rotate and crop the template image according to region properties
for j=1:num,
 
% Compute the coordinates for this region.
[x,y] = find(labels == j);
% Get an image that only has this region, the rest is black 
bwsegment = bwselect(aspectBW,y,x,8);
% Generate orignal gray scale with only one face region
oneface=immultiply(bwsegment,imgray);

%centroid of the region
cx1=centroids(j,1);
cy1=centroids(j,2);
%width and height of the region
p=regionprops(bwlabel(bwsegment),'BoundingBox');
boxdim=cat(1,p.BoundingBox);
regw=boxdim(3);
regh=boxdim(4);   
ratio=regh/regw;
%if region is too long, set to resonable new height and shift centroid up
if(ratio>1.6)
regh=1.5*regw;
cy1=cy1-(0.1*regh);        
end

%resize the model with same scale as the region
gmodel_resize=imresize(imtemplate,[regh regw],'bilinear');
%rotate the resized model by the angle of orientation of the region
if(angles(j)>0)
gmodel_rotate=imrotate(gmodel_resize,angles(j)-90,'bilinear','loose');
else
gmodel_rotate=imrotate(gmodel_resize,90+angles(j),'bilinear','loose');
end

%computing the centroid and size of the model
bwmodel=im2bw(gmodel_rotate,0);
%size of the model before crop
[g,h]=size(bwmodel);
%ensure that the bw model region has only one region
bwmorphed = bwmorph(bwmodel,'clean');
[L,no]=bwlabel(bwmorphed,8);
if(no==1)
bwsingle=bwmorphed;
else
ar=regionprops(bwlabel(bwmorphed),'Area');
areas=cat(1,ar.Area);  
[C,I]=max(areas);
% Compute the coordinates for this region.
[x1,y1] = find(bwlabel(bwmorphed)== I);
% Get an image that only has this region, the rest is black 
bwsingle = bwselect(bwmorphed,y1,x1,8);
end
%fill the model and crop - this option of region props crops automatically
filledmodel=regionprops(bwlabel(bwsingle),'FilledImage');
bwcrop=filledmodel.FilledImage;
%size of the scaled and rotated model after crop
[modh,modw]=size(bwcrop);
%crop the grayscale model to same size as that of bwlabel model
gmodel_crop=imresize(gmodel_rotate,[modh modw],'bilinear');
%centroid of scaled,rotated and cropped model
cenmod=regionprops(bwlabel(bwcrop),'Centroid');
central=cat(1,cenmod.Centroid);
cx2=central(1,1);
cy2=central(1,2);

mfit = zeros(size(oneface));
mfitbw = zeros(size(oneface));
[limy, limx] = size(mfit);
  
 % Compute the coordinates of where the face model is going to be in
 % the main image
 
 startx = cx1-cx2;
 starty = cy1-cy2;
 endx = startx + modw-1;
 endy = starty + modh-1;	  
   
  % Check for boundaries of the image
  startx = checklimit(startx,limx);
  starty = checklimit(starty,limy);
  endx = checklimit(endx,limx);
  endy = checklimit(endy,limy);
  
  % The following is to generate a new image having the same
  % size as the original one, but with the face of the model on it
  % rotated accordingly.
 
  for i=starty:endy,
   for j=startx:endx,
     mfit(round(i),round(j)) = gmodel_crop(round(i-starty+1),round(j-startx+1));
    end;
  end;

%  subplot(3,3,9);
%  imshow(mfit,[0 255])  
%      
% Get the cross-correlation value between model and region
gray_matched=gray_matched+mfit;
crosscorr =corr2(mfit,oneface);
%if cross-correlation value higher than threshold, add the region 
if(crosscorr>=0.6)    
    template_passed=template_passed+bwsegment;
    
end;
subplot(4,3,11);
imshow(gray_matched,[0 255])
title('Template Matching')
% subplot(4,3,12);
% imshow(template_passed)
end;

%**************************************************************************
%Following function return aspect ratios of regions in binary image
function [ratiolist] = get_aspect(inputBW)
%compute length of major axis (of ellipse)for each region
major = regionprops(inputBW,'MajorAxisLength');
major_length=cat(1,major.MajorAxisLength);
%compute length of minor axis (of ellipse)for each region
minor = regionprops(inputBW,'MinorAxisLength');
minor_length=cat(1,minor.MinorAxisLength);
%compute aspect ratio
ratiolist=major_length./minor_length;

%**************************************************************************
% Verifies that the coordinate is between the image region.
function newcoord = checklimit(coord,maxval)

  newcoord = coord;
  if (newcoord<1) newcoord=1; end;
  if (newcoord>maxval) newcoord=maxval; end;
  
%**************************************************************************

